/*
 * Copyright (c) 2016 Ian Sutton <ian@kremlin.cc>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF MIND, USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "tipru.h"

typedef enum { WRITE, CONTROL, NONE } strategy;
typedef enum { PRU_INSTR, PRU_DATA, PRU_SHARED } pru_space;

#define TARGET_DATA   (1 << 0)
#define TARGET_INSTR  (1 << 1)
#define TARGET_SHARED (1 << 2)

#define PRU0_DATA_OFFSET  0x00000
#define PRU1_DATA_OFFSET  0x02000
#define PRU0_INSTR_OFFSET 0x34000
#define PRU1_INSTR_OFFSET 0x38000
#define SHARED_OFFSET     0x10000

static void usage(void);
static int  pru_write(int, int, int, pru_space);

int
main(int argc, char *argv[])
{
	int prufd, infd_d, infd_i, infd_s, prunum, opt, error, runstate, ctl_send;
	uint8_t target;
	strategy strat;

	infd_d = infd_i = prufd = prunum = infd_s = -1;
	strat = NONE;
	target = 0x0;
	error = 0;

	if ((prufd = open("/dev/tipru", O_WRONLY)) == -1) {
	    fprintf(stderr,
		"Failed to open /dev/tipru for writing.\n");
	    goto err;
	}

	while ((opt = getopt(argc, argv, "t:s:n:d:i:r")) != -1) {
	    switch (opt) {
	    case 't':
		if (strat != NONE)
		    goto err;
		strat = CONTROL;
		runstate = atoi(optarg);
		break;

	    case 's':
		if (strat == CONTROL)
		    goto err;
		strat = WRITE;
		target |= TARGET_SHARED;
		if ((infd_s = open(optarg, O_RDONLY)) == -1) {
		    fprintf(stderr, "Could not open shared-space infile.\n");
		    goto err;
		}
		break;

	    case 'n':
		prunum = atoi(optarg);
		break;

	    case 'd':
		if (strat == CONTROL)
	            goto err;
		strat = WRITE;
		target |= TARGET_DATA;
		if ((infd_d = open(optarg, O_RDONLY)) == -1) {
		    fprintf(stderr, "Could not open data-space infile.\n");
		    goto err;
		}
		break;

	    case 'i':
		if (strat == CONTROL) 
	            goto err;
		strat = WRITE;
		target |= TARGET_INSTR;
		if ((infd_i = open(optarg, O_RDONLY)) == -1) {
		    fprintf(stderr, "Could not open instruction-space infile.\n");
		    goto err;
		}
	        break;
	    case 'r':
		if(strat != NONE)
		    goto err;
		ioctl(prufd, PRURESET);
		close(prufd);
		return 0;
		break;
	    case '?':
	    /* FALLTHROUGH */
	    default:
		goto err;
	    }
	}

	if (strat == NONE) {
	    goto err;
	} else if (prunum != 0 && prunum != 1) {
	    fprintf(stderr,
		"-n must specify either '1' or '0' for PRU cores 1/0 respectively.\n");
	    goto err;
	} else if (strat == CONTROL && runstate != 0 && runstate != 1) {
	    fprintf(stderr,
		"-t must specify either '1' or '0' for starting/stopping PRU respectively\n");
	}

	ctl_send = 0;
	if (strat == CONTROL) {
	    if (prunum)
		ctl_send |= PRUCTL_PRU1_SEL;
	    else
		ctl_send |= PRUCTL_PRU0_SEL;
	    if (runstate) {
		ctl_send |= PRUCTL_PRU0_EN;
		ctl_send |= PRUCTL_PRU1_EN;
	    }

	    ioctl(prufd, PRUCTL, &ctl_send);
	    close(prufd);
	    return 0;
	}

	if (target & TARGET_DATA) {
	    error |= pru_write(prufd, infd_d, prunum, PRU_DATA);
	    close(infd_d);
	} if (target & TARGET_INSTR) {
	    error |= pru_write(prufd, infd_i, prunum, PRU_INSTR);
	    close(infd_i);
	} if (target & TARGET_SHARED) {
	    error |= pru_write(prufd, infd_s, prunum, PRU_SHARED);
	    close(infd_s);
	}

	close(prufd);

	return error;

err:
	if (infd_d != -1)
	    close(infd_d);
	if (infd_i != -1)
	    close(infd_i);
	if (infd_s != -1)
	    close(infd_s);
	if (prufd != -1)
	    close(prufd);

	usage();
}

static void
usage()
{
	(void)fprintf(stderr,
	    "usage: pructl -n num -t state\n");
	(void)fprintf(stderr,
	    "usage: pructl -r\n");
	(void)fprintf(stderr,
	    "       pructl -n num [-d infile] [-i infile] [-s infile]\n");
	exit(1);
}

static int
pru_write(int prufd, int fd, int core_id, pru_space space)
{
	char *buf;
	struct stat *sb;
	ssize_t written;

	if (!(sb = (struct stat *) calloc(1, sizeof(struct stat))))
	    return -1;

	fstat(fd, sb);

	if (!(buf = (char *) calloc(1, sb->st_size)))
	    return -1;

	read(fd, buf, sb->st_size);

	if (space == PRU_SHARED)
	    written = pwrite(prufd, buf, (size_t)sb->st_size, SHARED_OFFSET);

	else if (core_id) {
	    if (space == PRU_DATA)
		written = pwrite(prufd, buf, (size_t)sb->st_size, PRU1_DATA_OFFSET);
	    else
		written = pwrite(prufd, buf, (size_t)sb->st_size, PRU1_INSTR_OFFSET);
	} else {
	    if (space == PRU_DATA)
		written = pwrite(prufd, buf, (size_t)sb->st_size, PRU0_DATA_OFFSET);
	    else
		written = pwrite(prufd, buf, (size_t)sb->st_size, PRU0_INSTR_OFFSET);
	}

	free(buf);
	free(sb);

	if (written == -1) {
	    fprintf(stderr, "Failed to write to device memory\n");
	    return -1;
	}

	return 0;
}
